#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/select.h>

#define BAUDRATE B115200
#define MAGICSTR "M4g1c!"

#include "config.h"

//This routine sends the magic string to init communications with the AVR.
void sendmagic(int f){
    char buff[]=MAGICSTR;
    write(f,buff,6);
}

//Send a char to the AVR
void sendchar(int f, unsigned char c) {
    write(f,&c,1);
    usleep(100000); 
}

//Read a char from the AVR. Time-out in 1 second, returns 255 on time-out.
unsigned char readchar(int f) {
    unsigned char buff[1];
    struct timeval tv;
    fd_set rfds;
    tv.tv_sec=1;
    tv.tv_usec=0;
    FD_ZERO(&rfds);
    FD_SET(f,&rfds);
    if (select(f+1,&rfds,NULL,NULL,&tv)){
	read(f,buff,1);
    } else {
	buff[0]=255;
    }
    return buff[0];
}

//Read a long integer from the AVR.
int readlong(int f) {
    int r=0;
    r=readchar(f);
    r+=readchar(f)<<8;
    r+=readchar(f)<<16;
    return r;
}

//Checks if a file exists
int fexists(char *fn) {
    struct stat statbuf;
    if (stat(fn,&statbuf)==0) return 1; else return 0;
}

//Resynchronize communications with the AVR.
void resync(int c) {
    struct timeval tv;
    fd_set rfds;
    int r;
    int repeat;
    do {
	repeat=0;
	do {
	    if (repeat) sendchar(c,'!');
	    repeat=1;
	    tv.tv_sec=1;
    	    tv.tv_usec=0;
	    FD_ZERO(&rfds);
	    FD_SET(c,&rfds);
	} while(!select(c+1,&rfds,NULL,NULL,&tv));
    } while (readchar(c)!='~');
}


//Main routine
int main(int argc, char** argv) {
    int comm,n,x,f;
    time_t tijd;
    struct stat statbuf;
    int r[4];
    int cf;
    char* env;
    unsigned char cal[4];
    unsigned char cmd[16];
    struct termios oldtio,newtio;

    //Open the port
    comm=open("/dev/ttyS0",O_RDWR);
    if (comm <0) {perror("Opening comport"); exit(-1); }

    //Set the correct baudrate etc
    tcgetattr(comm,&oldtio); // save current port settings
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag =  (CS8 | CREAD | CLOCAL | CRTSCTS);
    newtio.c_lflag = 0;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    cfsetospeed(&newtio,BAUDRATE);
    
    newtio.c_cc[VTIME]    = 0;   // inter-character timer unused 
    newtio.c_cc[VMIN]     = 1;   // blocking read until 1 chars received 
      
    tcflush(comm,TCIFLUSH);
    tcsetattr(comm,TCSANOW,&newtio);
    
    //Send the magic string
    sendmagic(comm);
    resync(comm);
    sendchar(comm,'T');
    for (x=0; x<4; x++) {
	char *env;
	char ename[16];
	sprintf(ename,"SENSCAL%i",x+1);
	env=getenv(ename);
#ifdef DEBUG
	printf("%s=%s\n",ename,env);
#endif
    	if (env!=NULL) {
	    cal[x]=atoi(env);	
	} else {
	    cal[x]=128;
	}
	sendchar(comm,cal[x]);
    }

    if (fexists(CMDFIFO)) {
	unlink(CMDFIFO);
    }
    mkfifo(CMDFIFO,0666);
    cf=open(CMDFIFO,O_RDONLY|O_NONBLOCK);

    //Main routine; 
    tijd=time(NULL);
    while(1) {
	tijd+=INTERVAL;
	while (tijd>time()){
	    //Check for any funny stuff happening to the time
	    if (tijd-time(NULL)<0 || tijd-time(NULL)>INTERVAL*5) tijd=time(NULL);

	    //Do we have a command passed in the fifo?
	    if ((x=read(cf,cmd,15))>0 && cmd[0]=='!') {
#ifdef DEBUG
		printf("Executing cmd.\n");
#endif
		cmd[x]=0; //make it zero-terminated
		//Go calibrate a channel
		if (cmd[1]=='C') {
		    int min,max,chan;
		    FILE *of;
		    resync(comm);
#ifdef DEBUG
		    printf("Calibrate start for chan %i.\n",cmd[2]-'1');
#endif
		    chan=cmd[2]-'1';
		    sendchar(comm,'C');
		    sendchar(comm,cmd[2]);
		    
		    //Wait till the user stops the calibration
		    while (read(cf,cmd,16)<=0) {
			usleep(10000);
		    }
		    sendchar(comm,'d'); //dummy
		    min=readchar(comm);
		    max=readchar(comm);
		    
		    //Write the calibration data for use by the webserver
		    of=fopen(CALOUTFILE,"w");
		    fprintf(of,"%i %i %i\n",chan+1,max,min);
		    fclose(of);
		    resync(comm);
		} else if(cmd[1]='T') {
		    int chan;
		    resync(comm);
		    //Read chan & val
		    chan=cmd[2]-'1';
		    cal[chan]=atoi((char *)&cmd[3]);
#ifdef DEBUG
		    printf("Set tresh: ch %i to %i.\n",chan+1,cal[chan]);
#endif
		    //And set the calibration in the AVR to the new value.
		    sendchar(comm,'T');
		    for (x=0; x<4; x++) sendchar(comm,cal[x]);
		    resync(comm);
		}
	    }
	    sleep(1);
	}
	//Interval has passed; read measurements from AVR and write to 
	//the data file
	resync(comm);

	sendchar(comm,'G');
	for (x=0; x<4; x++) r[x]=readlong(comm);
#ifdef DEBUG
	printf("%i %i %i %i\n",r[0],r[1],r[2],r[3]);
#endif
	f=open(DATFILE,O_WRONLY|O_CREAT|O_APPEND,0666);
	write(f,r,sizeof(int)*4);
	close(f);
    }
}




